;;;
;;; Virtual Disk Drive Project - July 19, 2005
;;;
;;;    Apple 2 "write" support.
;;;
;;;    Apple write runs in processor page 1.

#include "main.h"
#include "vd.h"
#include "utility.h"
#include "waitP1.mac"
#include "disk.h"
#include "apple.mac"
#include "apple.h"
#include "write.h"
#include "write.mac"
#include "nop.mac"
#include "sertrack.h"
#include "applewrite.mac"

#define CARRYBIT	STATUS,C
	
	global GCR6x2WriteSector
	global NIBWriteSector
		
PROG1	CODE
	
;;;*******************************************************************
;;; NAME:	TransSync()
;;;
;;; DESCR:	Syncronize to the transitions.  This routine knows
;;;		that it is called during the self-sync stream, but
;;;		really doesn't know when within the stream.  It finds
;;;		the transitions within the first part of the stream
;;;		and syncronizes as closely as it can to them.  It
;;;		returns with the concept that (0) is in the middle
;;;		of a cell.  The transition tracking registers are
;;;		set-up as well.  So following routines can simply
;;;		check for transitions at (0), or within a couple
;;;		cycles either way of (0).
;;;		NOTE - MemWriteMode is turned on here for convenience.
;;;
;;; ARGS:	
;;;
;;; RETURNS:	
;;;
;;; NOTES:	- called at (?) returns at (0) relative to the start
;;;		  of the transition.  The caller should think of this
;;;		  routine returning at (10) so that it will start
;;;		  sampling in the MIDDLE of a cell.
;;;		- we are at a low when called and when returned
;;;*******************************************************************
TransSync:
	;; we really don't know how long it took us to get into this
	;; routine after the WGATE was noticed.  That's OK though.
	;; We simply get ourselves syncronized as closely as possible
	;; to the transition edges, which we know occur every bit for
	;; the self sync bits.  Even if we land on the zeros at the
	;; end of a self-sync, though that probably won't happen, we'll
	;; be OK.

	SkipIfLow	WDATA	; (?) wait until we find a low
	GotoPrevious		; (?)

	;; at this point we know we've SEEN a low.  We may have
	;; seen it right at the beginning of its cycle, in the
	;; middle, or right at the end.  So now, wait for the
	;; low to terminate...

	SkipIfHigh	WDATA	; (?) wait until the end of the low
	GotoPrevious		; (?)

	;; Now we know we've found the end of the low, or I should
	;; say, the beginning of a high.  We know we're NEAR the
	;; beginning of the high, because we saw a low at the least
	;; 4 cycles ago.  So now we'll wait for the high to terminate.

	SkipIfLow	WDATA	; (?)
	GotoPrevious		; (?)

	;; Now we're narrowing in on the edge.  We know that we
	;; noticed the edge at either 0, 1, or 2, meaning that
	;; at this point we're at 2, 3, or 4.  However, we tend
	;; to "notice" edges 1 or 2 cycles late, so let's say
	;; that we're at 3, 4, or 5, and then split the difference
	;; saying we're at 4.
	
	;; since we're at a low, let's go ahead and set-up our sense logic

	TransSetup			; (4) sets-up all transition "things"

	;; turn on write mode as a nice service to our callers
	
	MemWriteMode			; (9) turn on write mode	
	
	;; so now we are (at best guess) at (13) in the middle of a low
	
	Nop5				; (13) to the start of the high

	return				; (18) return at (0) - the start of a high

;;;
;;; these are remapped registers which are used in GCR writing
;;;

#define IDATA		GCR_B1
#define DATACOUNTER	GCR_B2CLEAR
#define COUNTER		GCR_BYTECOUNT
#define LZFLAG		GCR_B2
	
;;;
;;; this needs to be the same as the register in apple.asm
;;; because NIB uses it.
;;; 
#define BYTE_COUNT	GCR_BYTECOUNT

;;; Note that RETDATA and RETDATA2 need to be adjacent registers
;;; with RETDATA2 after RETDATA.  They are accessed in one case
;;; through indirection.

#define RETDATA		GCR_OUT
#define RETDATA2	GCR_OUTREADY

;;;*******************************************************************
;;; NAME:	GCR6x2WriteSector()
;;;
;;; DESCR:	Used to write a 6x2 when the write gate is noticed.
;;;		It acts much like the read routine of the
;;;		Apple, and looks for the special indicator bytes for
;;;		the sector data.  Further, it ignores leading zeros
;;;		in the data.
;;;
;;; ARGS:	
;;;
;;; RETURNS:	
;;;
;;; NOTES:	- it is relatively unknown as to when this routine is
;;;		  called relative to the Raise of the WRITE_GATE and,
;;;		  therefore, the rise of the first bit of incoming
;;;		  data.  Further, the PulseSync routine may miss the
;;;		  incoming bit (NEED TO CHECK THE PULSE WIDTH OF THE
;;;		  APPLE WRITE PULSES)
;;;		- this routine obeys the "very odd" packing of the data
;;;		  into the SVD sectors, including the 4 bits in the
;;;		  header block.
;;;*******************************************************************
#define GETBYTE		GCRWS_GetByte
#define GETRAWBYTE	GCRWS_GetRawByte

#define BYTESYNC	GCRWS_ByteSync
#define HEADER		GCRWS_Header

#define DATA		GCRWS_Data
#define DATA_2bit_phase	GCRWS_Data2
#define DATA_4bit_phase	GCRWS_Data4
#define DATA_6bit_phase	GCRWS_Data6
#define DATA_0bit_phase	GCRWS_Data0
#define EPILOG		GCRWS_Epilog
#define CHECKSUM	GCRWS_Checksum

GCR6x2WriteSector:	
	call	TransSync		; (?) turns on MemWriteMode too
	Nop4				; (10)

	;; set up the data counter and bit accumulator for the loop below
	
BYTESYNC:	

	;; rather useless to just keeping setting DATACOUNTER in this loop,
	;; but the timing worked out well...
	
	movlw	d'86'			; (14) plan to loop 86 times, though the first time is only partial
	movwf	DATACOUNTER		; (15)
	
	;; now allow the incoming self-syncs to get us into alignment, we look for a D5

	SkipIfHigh	WRITE_GATE	; (16) skips straight to HEADER if WRITE_GATE falls
	goto		HEADER		; (17)   we'll get useless data, but at least we won't lock-up
	
	call	GETRAWBYTE		; (18) called on schedule (0)
	sublw	0xd5			; (10) look for the start of the prolog
	skpz				; (11) if zero, we've hit our D5
	goto	BYTESYNC		; (12)   got something else, loop back around

HEADER:
	movlw	0xd5			; (13) write the start of the prolog to memory
	MemWriteLowInc			; (14)	

	Nop2				; (16)
	
	;; note that we're calling the byte routines a bit early here, this is to
	;; give us time later.  Note that calling them early isn't really a problem
	;; due the "transition" nature of the apple signal.  Though I try to keep
	;; it within a few cycles of "zero"
	
	call	GETRAWBYTE		; (18) called on schedule (0)
	MemWriteLowInc			; (10)
	Nop6				; (12)

	call	GETRAWBYTE		; (18) on schedule (+0)
	MemWriteLowInc			; (10)
	Nop5				; (12)
	
DATA:		
	;; now we read the first byte of data, which is treated specially
	;; the top 4 bits are loaded into the header, then we shift to the
	;; data section
	
	call	GETBYTE			; (17) called one cycle early (-1)
	rlf	INDF			; (8) before this, we have |xx12 3456| where x is zero
	rlf	INDF			; (9) now have |1234 56xx| where x is zero (counts on GETBYTE clearing carry)
	swapf	INDF,W			; (10) now have |56xx 1234|
	movwf	MEMDATA			; (11) prepare to write it
	MemWriteLowInc			; (12) top bits aren't zero, but really should be (apple.asm doesn't break tho)

	movlw	0xc0			; (13)
	andwf	MEMDATA			; (14) clears all but the top two bits of MEMDATA which was already set
	
	MemAddrInc	SECCOUNT	; (15) flip to data block
	
	;; this is real weird, but the top bits in MEMDATA aren't zero like they should be.
	;; The header header "preload" byte should only have 4 bits in the lower nibble.
	;; However the code in the reader part effectively ignores the upper bits if they
	;; are present.  The nice part, is at this point, the two top bits are exactly
	;; where they should be the the rest of this routine.
		
	;; at this point we need to read 341 bytes
	;; Since we have 4 phases to go through, we do the loop 85 (/ 340 4) full times
	;;  and one extra time for just the first 2-bit phase ("DATACOUNTER" prepared above)

#define LEFTOVERBITS	MEMDATA

	goto	DATA_2bit_phase		; (17)

DATA_0bit_phase:	;;  - have have no bits left over, so the incoming data is just stored

	Nop1				; (17)
	call	GETBYTE			; (18) called on schedule (0)
	movf	INDF,W			; (9)
	movwf	LEFTOVERBITS		; (10) data is right-justified in LEFTOVERBITS

	Nop6				; (11)

DATA_6bit_phase:	;; - have 6 bits left over, need 2 more -  LEFTOVERBITS is right-justified

	call	GETBYTE			; (17) called 1 cycle early (-1)
	rlf	INDF			; (8) get RETDATA left justified
	rlf	INDF			; (9)  (counts on GetByte to clear carry)
	rlf	INDF			; (10) now shift new data into left over data (carry WILL have zero)
	rlf	LEFTOVERBITS		; (11)
	rlf	INDF			; (12)
	rlf	LEFTOVERBITS		; (13) this is the byte to write

	MemWriteLowInc			; (14)

	movf	INDF,W			; (16) this is now left-justified with 4 left-over bits
	movwf	LEFTOVERBITS		; (17)

DATA_4bit_phase:	;; - there are 4 bits left justified in LEFTOVERBITS

	;; need to take the top four bits of the incoming data, and add them to the existing data
	;; but the incoming data is right-justified...quite the bummer, so we get tricky with "swap"

	call	GETBYTE			; (18) called on schedule (+0)
	rlf	INDF			; (9) get RETDATA left justified
	rlf	INDF			; (10)  (counts on GetByte to clear carry)
	swapf	INDF,W			; (11) RETDATA is left justified, and W has swapped RETDATA
	andlw	0x0f			; (12) mask off other bits, ready to OR with LEFTOVERBITS
	iorwf	LEFTOVERBITS		; (13) this is the byte to write

	MemWriteLowInc			; (14)

	swapf	INDF,W			; (16)
	andlw	0xc0			; (17) now have the remaining two bits from RETDATA left-justified
	movwf	LEFTOVERBITS		; (18)

DATA_2bit_phase:	;; - we have 2 bits left over, need to read another byte
	
	;; since we know that LEFTOVERBITS has two high bits, with the others zero,
	;; we can simply OR them with incoming byte to create the byte to write

	call	GETBYTE			; (19) called one cycle late (+1)
	movf	INDF,W			; (10)
	iorwf	LEFTOVERBITS		; (11) this is the byte to write

	MemWriteLowInc			; (12)

	decfsz	DATACOUNTER		; (14) loop around for more bytes if needed
	goto	DATA_0bit_phase		; (15)
	
CHECKSUM:

	MemAddrLoSet			; (16) flip back to header block for Checksum
	call	GETRAWBYTE		; (18) called on schedule (+0)
	MemWriteLowInc			; (10)

EPILOG:		;; we don't do anything with the epilog, we just read it...

	Nop6				; (12)

	call	GETRAWBYTE		; (18) get an epilog byte
	Nop8				; (10)

	call	GETRAWBYTE		; (18) get an epilog byte
	Nop8				; (10)

	call	GETRAWBYTE		; (18) get an epilog byte

	Nop1				; (10)
	MemReadMode			; (11) we are done with the sector!
	goto	SerTrackSecDone		; (16)
		
;;;*******************************************************************
;;; NAME:	GCRWS_GetByte()
;;;
;;; DESCR:	Get an incoming byte and put it into W.  By the time
;;;		the byte is read, the translation into 6 bits has
;;;		already been done.
;;;
;;; ARGS:	
;;;
;;; RETURNS:	INDF - accessing this will cause access to either RETDATA
;;;			or RETDATA2.
;;;			RETDATA/RETDATA2 have 6 bits (right justified) of
;;;			GCR-mapped data	W is NOT set to a useful value.
;;;
;;; NOTES:	- some routines count on the carry being zero after call,
;;;		  and this routine does that
;;;		- although the first bit of a GCR byte is ALWAYS one,
;;;		  this routine maps that one to zero because the reverse
;;;		  GCR mapping table assumes a zero in the upper bit
;;;		  (though we all know that this really means a one)
;;;		- called at (18) returns at (9) (yikes)
;;;		- routine deviates (-1) and (+1) from (0)
;;;		- note that W DOES NOT have the same value as RETDATA
;;;		- assumes that WSENSE is setup right
;;;		- thinks that (0) is in the middle of the cell
;;;*******************************************************************

#define LEADING_ZEROS	GCRWSGB1
#define PROCESS_ZERO	GCRWSGB2

;;; GetRawBit - used to make better use of code space while making the code readable
;;;	This routine checks for a bit, and shift it into the incoming data accumulator:	  IDATA.
;;;	Uses 10 cycles, because that makes the below code use less space (fewer Nop's).

GetRawBit:				; called at (-2,18)
	TransSkipOnZero			; (0)
	bsf		IDATA,0		; (4)
	TransFlipSense			; (5)
	rlf		IDATA		; (6)
	
	;; this code will help us stay "in sync" with the incoming bit-stream
	;; a poor man's "pll" in a sense.  We check a bit early for a transition
	;; and if we see one,
	;; we adjust our sense of where we are, essentially dropping cycles,
	;; allowing us to "catch-up".
	;; 
	;; It is best to drop 2 cycles, as opposed to just 1.  Because we don't
	;; check at every bit, we can easily get "too behind" for just a single
	;; cycle "catch-up".  A 2 cycles catch-up works much better as shown
	;; by experimentation.
	;; 
	;; Note that this isn't perfect in that if we have an upcoming zero bit
	;; we won't see a transition and can't sync.  But then, again, the Apple
	;; scheme ensures that we don't have more than 2 zeros together, for just
	;; this reason.  Normal, in-sync, operation causes this routine to return
	;; at (16).  An adjustment will cause it to use one less cycle.

	;; note that in the following code, we are CHECKING for a transition
	;; that occurred earlier than we thought it would, so we adjust our
	;; position ahead.  Note, too, that we DON'T TransFlipSense when we
	;; see this, because we're not actually "consuming" the bit...we'll do
	;; that later.

 	Nop1				; (7)
	TransSkipOnZero			; (8)
	goto	GetRawBitRet		; (12) on a one skip forward to return
	Nop3				; (13)

GetRawBitRet:	
	return				; (14,16) returns at 16 when moving forward, 18 normally



GCRWS_GetByte:

LEADING_ZEROS:	
	TransSkipNonZero		; (0) load W with transition data
	goto		PROCESS_ZERO	; (4) if we have a zero process it below
	TransFlipSense			; (5) got the first bit, now flip the sense

	;; got leading 1, start the processing of the incoming bits
	;; again, note that the 1 is NOT placed in the incoming data byte,
	;; it is assumed in the GCR reverse mapping.

	clrf		IDATA		; (6) clears the data accumulator
	bcf		CARRYBIT	; (7) clear the carry bit too - 'cause it will be shifted in
	
	movlw		RETDATA		; (8) get the address of the return data register into 
	movwf		FSR		; (9)    the indirect register for later.

	Nop8				; (10)
	
	call		GetRawBit	; (18) called on schedule - bit 2
	call		GetRawBit	; (18) called on schedule - bit 3
	call		GetRawBit	; (18) called on schedule - bit 4
	call		GetRawBit	; (18) called on schedule - bit 5
	call		GetRawBit	; (18) called on schedule - bit 6

	rlf		IDATA		; (18) go ahead and prepare for bits 7 and 8 ahead of time

	TransSkipOnZero			; (19) 1 cycle ahead of time (-1) - bit 7
	bsf		IDATA,1		; (3) sets bit one as opposed to zero for bit 7
	TransFlipSense			; (4) flip the sense (only effective on a one)

	;; pre-emptively map the incoming data from GCR - first do as if bit 8 is a zero
	
	movf		IDATA,W		; (5) go ahead and map what we got
	call		rGCR6x2MapUnsafe; (6) we KNOW we haven't set the top bit, so this is safe
	movwf		RETDATA2	; (12) save mapping as if bit 8 was a zero

	;; now map as if bit 8 is a one
	
	incf		IDATA,W		; (13) sets the low bit for mapping
	call		rGCR6x2MapUnsafe; (14) again, top bit not set so this is safe
	movwf		RETDATA		; (0) save mapping as if bit 8 was a one

	TransSkipNonZero		; (1) 1 cycle late (+1) - index points to RETDATA - bit 8
	incf		FSR		; (5) index now points to RETDATA2
	TransFlipSense			; (6) flip the sense (only effective on a one)
	return				; (7) now access to INDF will grab the right return data

PROCESS_ZERO:

	;; note that no sense-flipping is done here...and it shouldn't be because
	;; this bit was a zero - NOTE, however, that the TransFlipSense macro
	;; COULD be called, because it is only effective when the last bit was
	;; a one

	;; BIG NOTE - if we get stuck with WDATA with no transitions, as occurs when
	;;   the write gate drops, then we could loop forever looking for non-zero.
	;;   So we build in a safety measure such that when we see the write-gate go
	;;   down, we just return with random data.

 	Nop11				; (6)
	
	SkipIfLow	WRITE_GATE	; (17)
	goto		LEADING_ZEROS	; (18) loop around for another bit
	return				; (19) lost our WRITE_GATE, ABORT!  ABORT! - random data returned
	

;;;**********************************************************************
;;; NAME:	GetRawByte()
;;;
;;; DESCR:	Get a raw byte of "transition" oriented write data.
;;;		Much like GCRWS_GetByte, gets a byte from the write
;;;		line.  However, this routine gets a raw byte as opposed
;;;		a GCR-mapped byte.
;;;
;;; ARGS:	
;;;
;;; RETURNS:	RETDATA and W are set to the same thing
;;;
;;; NOTES:	- called at (18) returns at (12)
;;;		- this routine maintains zero deviation from (0)
;;;		- assumes that WSENSE is setup right
;;;		- thinks that (0) is in the middle of the cell
;;;**********************************************************************
#define RLEADING_ZEROS	GCRWSGRB1
#define RPROCESS_ZERO	GCRWSGRB2

#undefine RETDATA
#define RETDATA	MEMDATA
	
GCRWS_GetRawByte:

RLEADING_ZEROS:	
	TransSkipNonZero		; (0) load W with transition data
	goto		RPROCESS_ZERO	; (4) process an incoming zero
	TransFlipSense			; (5) flip the sense

	;; got leading 1, start the processing of the incoming bits
	;; again, note that the 1 is NOT placed in the incoming data byte yet,
	;; it is done later

	clrf		IDATA		; (6) clears the data accumulator
	bcf		CARRYBIT	; (7) clear the carry bit too - 'cause it will be shifted in

	Nop10				; (8)
	
	call		GetRawBit	; (18) called on schedule - bit 2
	call		GetRawBit	; (18) called on schedule - bit 3
	call		GetRawBit	; (18) called on schedule - bit 4
	call		GetRawBit	; (18) called on schedule - bit 5
	call		GetRawBit	; (18) called on schedule - bit 6
	call		GetRawBit	; (18) called on schedule - bit 7

	movf		IDATA,W		; (18) load up RETDATA
	movwf		RETDATA		; (19)

	TransSkipOnZero			; (0) on schedule - bit 8
	bsf		RETDATA,0	; (4) a transition equals a one
	TransFlipSense			; (5) flip the sense (only effective on a one)
	
	bsf		RETDATA,7	; (6) top bit needs to be one in raw read
	movf		RETDATA,W	; (7) loads the W with same as RETDATA
	return				; (8) returns at (10)

RPROCESS_ZERO:

	;; note that no sense-flipping is done here...and it shouldn't be because
	;; this bit was a zero - NOTE, however, that the TransFlipSense macro
	;; COULD be called, because it is only effective when the last bit was
	;; a one

	;; BIG NOTE - if we get stuck with WDATA with no transitions, as occurs when
	;;   the write gate drops, then we could loop forever looking for non-zero.
	;;   So we build in a safety measure such that when we see the write-gate go
	;;   down, we just return with random data.

 	Nop11				; (6)
	
	SkipIfLow	WRITE_GATE	; (17)
	goto		RLEADING_ZEROS	; (18) loop around for another bit
	return				; (19) lost our WRITE_GATE, ABORT!  ABORT! - random data returned
	
;;;*******************************************************************
;;; NAME:	NIBWriteSector()
;;;
;;; DESCR:	Though this is labeled as "write sector" this routine
;;;		actually continously writes an entire track, much like
;;;		its analogous routine NIBSector.
;;;
;;;		The NIBSector routine jumps to this routine the moment
;;;		a WRITE is noticed.  This routine "picks up" where the
;;;		NIBSector routine left off, adopting the same concept
;;;		of the memory pointer, and track/block pointer.
;;;
;;;		It will return only when the write line is lowered,
;;;		OR the disk selection lines go down.  Note that we can't
;;;		anticipate when the write line will be lowered...and
;;;		it will probably be in the middle of a block of data.
;;;		So great pains are taken to ensure that write is stopped
;;;		immediately so that other bytes won't get overwritten
;;;		within a block.
;;; 
;;;		After return, the disk "keeps spinning" (generating READ
;;;		data) until the end of the track  notices that the disk
;;;		lines have gone down.  It is, therefore, the READ routines
;;;		that align the memory pointers correctly.
;;; 
;;;
;;; ARGS:	BYTE_COUNT tracks the number of bytes in the sector
;;;		SECCOUNT tracks the sector within the track we're working on
;;;		 - great care will have to be taken to ensure that update
;;;		   these two variables appropriately given the location in
;;;		   which we notice the write.
;;;		 - these two variables need to be maintained as if we were
;;;		   generating data.
;;;
;;; RETURNS:	
;;;
;;; NOTES:	
;;;*******************************************************************
#define	SYNC_ON_ZERO	NIBWS_SYNCZERO
#define START_LOOP	NIBWS_STARTUP
#define WRITE_LOOP	NIBWS_WRITELOOP
#define NIB_END		NIBWS_NIBEND
#define NIB_PROC_ZERO	NIBWS_NIBPROCESSZERO
#define BIT0		NIBWS_BIT0
			
NIBWriteSector:
	;; called when a "write" is noticed

	call	TransSync		; (?) turns on MemWriteMode too

	Nop10				; (10)
	
	;; at this point we are as "in-sync" as we're going to get
	;; and unlike with writing of GCR sectors, we go a little
	;; farther and consume the first zero leading up to the next
	;; byte so we don't write NON self syncs when they are
	;; being written from the computer during start of write

SYNC_ON_ZERO:	
	TransSkipNonZero		; (0) load W with transition data, looking for a zero
	goto		START_UP	; (4) process an incoming zero
	TransFlipSense			; (5) flip the sense

	;; potential lock-up condition here, if a zero never comes...

	Nop12				; (6)
	goto		SYNC_ON_ZERO	; (18) still looking for a zero

START_UP:

	movlw		0xff		; (6) set up to write this initial self-sync
	movwf		MEMDATA		; (7)
	Nop5				; (8)

	;; at this point we have read the first zero of a two-zero self-sync
	;; so the next bit WILL be zero.

WRITE_LOOP:

	Nop4				; (13)
	bsf		LZFLAG,0	; (17) a zero in LZFLAG indicates a leading zero (this "clears" the flag)
	movlw		0x02		; (18) go ahead and prepare IDATA
	movwf		IDATA		; (19)   with the first bit shifted in

BIT0:	
;; bit0 - do it manually
	
	TransSkipNonZero		; (0) load W with transition data
	goto		NIB_PROC_ZERO	; (4) process an incoming zero (returns to bit0)
	TransFlipSense			; (5) flip the sense

	;; deal with a potential self-sync from the LAST byte
	;; if the last byte was an 0xff  AND  we had a leading zero this time (LZFLAG is zero)
	;; then reset the last byte to the self-sync token (zero)  before writing it
	
	incf		MEMDATA,W	; (6) sets Z if last byte was an 0xff
	skpnz				; (7)
	movf		LZFLAG		; (8) if LZFLAG is zero...
	skpnz				; (9)
	clrf		MEMDATA		; (10) MEMDATA now has the self-sync token

	MemWriteLowInc			; (11) write out the last byte and increment (mungs carry)

	bcf		CARRYBIT	; (13) clear the carry bit - 'cause it will be shifted in

	Nop4				; (14)
	
;; bit1
	call	GetRawBit		; (18) get the next bit into IDATA - Bit 1
;; bit2
	call	GetRawBit		; (18) get the next bit into IDATA - Bit 2
;; bit3
	call	GetRawBit		; (18) get the next bit into IDATA - Bit 3

	Nop1				; (18)
	decf		BYTE_COUNT	; (19) prepare for next byte
	
;; bit4 - do it manually
	
	TransSkipOnZero			; (0) load W with transition data
	bsf		IDATA,0		; (4)
	TransFlipSense			; (5) flip the sense
	rlf		IDATA		; (6)
		
	;; this confusing piece of code says (1) check to see if SECCOUNT equals
	;; SECTORS (ie - we're done with a track), (2) if so, then check to see
	;; if the BYTE_COUNT is zero (ie - we're done with the sector), if we are
	;; then (3) reset the sector count.

	movf		SECCOUNT,W	; (7) if on 26, and done with the sector then
	subwf		SECTORS,W	; (8)      we're done with the track
	skpnz				; (9) if NOT done with track, then double skip
	movf		BYTE_COUNT	; (10) resets the Z flag based upon BYTE_COUNT
	skpnz				; (11) if NOT end of sector then skip
	clrf		SECCOUNT	; (12) if end of track, reset sector counter
	bcf		CARRYBIT	; (13) above code messed with the carry bit

	;; increment the sector count if we are at the end of the sector
	;; moves us to either the next sector or sector 1

	movf		BYTE_COUNT	; (14) 
	skpnz				; (15)
	incf		SECCOUNT	; (16)
	
	Nop1				; (17)

;; bit5
	call	GetRawBit		; (18) get the next bit into IDATA - Bit 5
;; bit6
	call	GetRawBit		; (18) get the next bit into IDATA - Bit 6

;; bit7 - do it manually

	MemAddrLoSet			; (18) reset the base for later
	
	TransSkipOnZero			; (0) load W with transition data
	bsf		IDATA,0		; (4) 
	TransFlipSense			; (5) flip the sense
		
	MemAddrInc	SECCOUNT	; (6) adjust for target sector

	;; IDATA now has the whole byte in it, it will be written
	;; on the next loop - or if we are done writing

	movf		IDATA,W		; (8)
	movwf		MEMDATA		; (9)

	SkipIfLow	WRITE_GATE	; (10)
	goto		WRITE_LOOP	; (11)

NIB_END:
	;; write the current byte out, but don't increment, let the caller do that

	MemWriteByte			; () don't worry that we may have had a self sync

	;; at this point, the BYTE_COUNT/SECCOUNT hasn't been updated
	;; count on the caller to do that

	;; 	SHOULD THE MEMWRITEMODE AND MEMREADMODE BE CALLS?
	;; 	IT WOULD TAKE LONGER, BUT NORMALLY WE HAVE MORE TIME
	;; 
	;; maybe someday, when we need the space
	
	;; note that we are, by definition, out of sync with our previous
	;; "read" data generation, so we really don't worry about timing here

	MemReadMode
	return				; unlike in other write routines, this one 
					;   is CALLed. 
	
NIB_PROC_ZERO:

 	Nop10				; (6)

	;; the WRITE_GATE is checked here because a drop in it could cause
	;; us to loop indefinitely

	clrf		LZFLAG		; (16) set the "leading zero" flag
	SkipIfLow	WRITE_GATE	; (17)
	goto		BIT0		; (18) loop around for another bit
	goto		NIB_END		; (19) lost our WRITE_GATE so clean-up
		
;;; =======================================================================
;;; =======================================================================

	END


